import importlib.resources
import json
import socketserver
import traceback
import webbrowser
from abc import ABC, abstractmethod
from base64 import b64encode
from collections import OrderedDict
from dataclasses import dataclass
from http import HTTPStatus
from http.server import SimpleHTTPRequestHandler
from typing import Any, Optional
from typing import OrderedDict as OrderedDictType
from urllib.parse import parse_qs, urlparse

from solders.keypair import Keypair

__all__ = [
    "SignedMessage",
    "Signer",
    "KeypairSigner",
    "create_message",
]


@dataclass(kw_only=True)
class SignedMessage:
    message_b64: str
    wallet_pubkey_b58: str
    signature_b64: str

    def to_dict(self):
        return dict(self.__dict__)


def _base64_encode(data: bytes):
    return b64encode(data).decode("ascii")


class Signer(ABC):

    def sign_json(
        self,
        message: Any,
    ):
        json_message = json.dumps(message).encode("utf-8")
        signed_message = self.sign_raw(json_message)

        return signed_message

    @abstractmethod
    def sign_raw(
        self,
        message: bytes,
    ) -> SignedMessage:
        ...


class BrowserExtensionSigner(Signer):

    def sign_raw(
        self,
        message: bytes,
    ) -> SignedMessage:
        running = True

        public_key: Optional[str] = None
        signature: Optional[str] = None

        html_template = importlib.resources.read_text(__package__, "web_sign.html")  # type: ignore

        class Handler(SimpleHTTPRequestHandler):
            def do_GET(self):
                try:
                    nonlocal running, public_key, signature

                    parsed_path = urlparse(self.path)

                    params = parse_qs(parsed_path.query)

                    signature_param = params.get("signature")
                    if signature_param is not None:
                        signature = signature_param[0]
                    else:
                        signature = None

                    public_key_param = params.get("publicKey")
                    if public_key_param is not None:
                        public_key = public_key_param[0]
                    else:
                        public_key = None

                    if parsed_path.path == "/result" and signature and public_key:
                        running = False
                        self.send_html("OK")
                    else:
                        content = html_template.replace('"%MESSAGE_FROM_PYTHON%"', str([int(x) for x in message]))
                        self.send_html(content)
                except:
                    traceback.print_exc()

            def send_html(self, content: str):
                self.send_response(HTTPStatus.OK)
                self.send_header("Content-Length", str(len(content)))
                self.send_header("Content-Type", "text/html")
                self.end_headers()

                self.wfile.write(content.encode())

        port = self._find_open_port()

        with socketserver.TCPServer(("", port), Handler) as httpd:
            url = f"http://localhost:{port}"
            print(f"sign: open {url} in your browser")
            webbrowser.open_new_tab(url)

            while running:
                httpd.handle_request()

        assert public_key
        assert signature

        return SignedMessage(
            message_b64=_base64_encode(message),
            wallet_pubkey_b58=public_key,
            signature_b64=signature,
        )

    def _find_open_port(self):
        import socket

        with socket.socket() as socket:
            socket.bind(("", 0))
            return socket.getsockname()[1]


class KeypairSigner(Signer):

    def __init__(self, wallet: Keypair):
        self._wallet = wallet

    def sign_raw(
        self,
        message: bytes,
    ) -> SignedMessage:
        signature = self._wallet.sign_message(message)

        return SignedMessage(
            message_b64=_base64_encode(message),
            wallet_pubkey_b58=str(self._wallet.pubkey()),
            signature_b64=_base64_encode(bytes(signature)),
        )

    @staticmethod
    def load(wallet_path: str):
        with open(wallet_path) as file:
            wallet_data = file.read()

        wallet = Keypair.from_json(wallet_data)

        return KeypairSigner(wallet)


def create_message(
    *,
    cert_pub: str,
    hotkey: str,
    model_id: str,
) -> OrderedDictType[str, str]:
    message: OrderedDictType[str, str] = OrderedDict()
    message["cert_pub"] = cert_pub
    message["hotkey"] = hotkey
    message["model_id"] = model_id

    return message



import os
from typing import Literal, Optional, get_args

import click

import crunch_certificate.certificate as certificate
import crunch_certificate.pem as pem
import crunch_certificate.sign as sign
from crunch_certificate.__version__ import __version__


@click.group()
@click.version_option(__version__, package_name="__version__.__title__")
def cli():
    pass  # pragma: no cover


@cli.group(name="ca")
def ca_group():
    pass  # pragma: no cover


@ca_group.command(name="generate")
@click.option("--common-name", type=str, required=True, prompt=True)
@click.option("--organization-name", type=str, required=True, prompt=True)
@click.option("--key-path", type=click.Path(dir_okay=False, writable=True), default="ca.key", prompt=True)
@click.option("--cert-path", type=click.Path(dir_okay=False, writable=True), default="ca.crt", prompt=True)
@click.option("--overwrite", is_flag=True)
def ca_generate(
    common_name: str,
    organization_name: str,
    key_path: str,
    cert_path: str,
    overwrite: int,
):
    if os.path.exists(key_path) and not overwrite:
        click.echo(f"{key_path}: file already exists (bypass using --overwrite)", err=True)
        raise click.Abort()

    if os.path.exists(cert_path) and not overwrite:
        click.echo(f"{cert_path}: file already exists (bypass using --overwrite)", err=True)
        raise click.Abort()

    (
        ca_key,
        ca_cert,
    ) = certificate.generate_ca(
        common_name=common_name,
        organization_name=organization_name,
    )

    ca_key_pem = pem.dumps(private_key=ca_key)
    ca_cert_pem = pem.dumps(certificate=ca_cert)

    with open(key_path, "w") as fd:
        fd.write(ca_key_pem)
    click.echo(f"ca: {key_path}: saved key")

    with open(cert_path, "w") as fd:
        fd.write(ca_cert_pem)
    click.echo(f"ca: {cert_path}: saved certificate")


@cli.group(name="tls")
def tls_group():
    pass  # pragma: no cover


TargetProfileString = Literal["coordinator", "cruncher"]


@tls_group.command(name="generate")
@click.option("--ca-key-path", type=click.Path(dir_okay=False, readable=True, exists=True), default="ca.key")
@click.option("--ca-cert-path", type=click.Path(dir_okay=False, readable=True, exists=True), default="ca.crt")
@click.option("--common-name", type=str, required=True, prompt=True)
@click.option("--target", type=click.Choice(get_args(TargetProfileString)), required=False)
@click.option("--key-path", type=click.Path(dir_okay=False, writable=True), default="tls.key", prompt=True)
@click.option("--cert-path", type=click.Path(dir_okay=False, writable=True), default="tls.crt", prompt=True)
@click.option("--overwrite", is_flag=True)
def tls_generate(
    ca_key_path: str,
    ca_cert_path: str,
    common_name: str,
    target: Optional[TargetProfileString],
    key_path: str,
    cert_path: str,
    overwrite: int,
):
    if os.path.exists(key_path) and not overwrite:
        click.echo(f"{key_path}: file already exists (bypass using --overwrite)", err=True)
        raise click.Abort()

    if os.path.exists(cert_path) and not overwrite:
        click.echo(f"{cert_path}: file already exists (bypass using --overwrite)", err=True)
        raise click.Abort()

    with open(ca_key_path) as fd:
        ca_key = pem.loads_private_key(fd.read())
    click.echo(f"ca: {ca_key_path}: loaded key")

    with open(ca_cert_path) as fd:
        ca_cert = pem.loads_certificate(fd.read())
    click.echo(f"ca: {ca_cert_path}: loaded certificate")

    if target == "coordinator":
        is_client = True
        is_server = False
    elif target == "cruncher":
        is_client = True
        is_server = False
    else:
        is_client = True
        is_server = True

    (
        tls_key,
        tls_cert,
    ) = certificate.generate_tls(
        ca_key=ca_key,
        ca_cert=ca_cert,
        common_name=common_name,
        is_client=is_client,
        is_server=is_server,
    )

    tls_key_pem = pem.dumps(private_key=tls_key)
    tls_cert_pem = pem.dumps(certificate=tls_cert)

    with open(key_path, "w") as fd:
        fd.write(tls_key_pem)
    click.echo(f"tls: {key_path}: saved key")

    with open(cert_path, "w") as fd:
        fd.write(tls_cert_pem)
    click.echo(f"tls: {cert_path}: saved certificate")


@cli.command(name="sign")
@click.option("--tls-cert-path", type=click.Path(dir_okay=False, readable=True, exists=True), default="tls.crt")
@click.option("--hotkey", type=str, required=True)
@click.option("--model-id", type=str, required=True)
@click.option("--tls-cert-path", type=click.Path(dir_okay=False, readable=True, exists=True), default="tls.crt")
@click.option("--wallet-path", type=click.Path(dir_okay=False, readable=True, exists=True), required=False)
def sign_command(
    tls_cert_path: str,
    hotkey: str,
    model_id: str,
    wallet_path: Optional[str],
):
    with open(tls_cert_path) as fd:
        tls_cert = pem.loads_certificate(fd.read())
    click.echo(f"tls: {tls_cert_path}: loaded certificate")

    # TODO Use real values
    message = sign.create_message(
        cert_pub=certificate.get_public_key_as_string(tls_cert),
        hotkey=hotkey,
        model_id=model_id,
    )

    if wallet_path is not None:
        signer = sign.KeypairSigner.load(wallet_path)
    else:
        signer = sign.BrowserExtensionSigner()
    
    signed_message = signer.sign_json(message)

    print(f"message_b64       : {signed_message.message_b64}")
    print(f"wallet_pubkey_b58 : {signed_message.wallet_pubkey_b58}")
    print(f"signature_b64     : {signed_message.signature_b64}")



__title__ = "crunch-certificate"
__description__ = "crunch-certificate - Certificate module for the CrunchDAO Platform"
__version__ = "0.1.0"
__author__ = "Enzo CACERES"
__author_email__ = "enzo.caceres@crunchdao.com"
__url__ = "https://github.com/crunchdao/crunch-certificate"



"""
crunch-certificate
~~~~~~~~~~~~~~~~~~
The crunch module to generate and use certificates!
"""

import crunch_certificate.certificate as certificate  # type: ignore
import crunch_certificate.pem as pem  # type: ignore
import crunch_certificate.private_key as private_key  # type: ignore
import crunch_certificate.sign as sign  # type: ignore
from crunch_certificate.__version__ import __version__



from base64 import b64encode
from datetime import datetime, timedelta, timezone
from typing import List, Tuple

from cryptography import x509
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.serialization import Encoding, PublicFormat
from cryptography.x509.oid import ExtendedKeyUsageOID, NameOID, ObjectIdentifier

from crunch_certificate.private_key import PrivateKey
from crunch_certificate.private_key import generate as generate_private_key

__all__ = [
    "generate_ca",
    "generate_tls",
    "DEFAULT_DAYS_VALID",
]

DEFAULT_DAYS_VALID = 99 * 365


def generate_ca(
    *,
    common_name: str,
    organization_name: str,
    days_valid: int = DEFAULT_DAYS_VALID,
) -> Tuple[
    PrivateKey,
    x509.Certificate,
]:
    ca_key = generate_private_key(
        type="rsa",
    )

    # Build subject/issuer name (self-signed CA)
    subject = issuer = x509.Name([
        x509.NameAttribute(NameOID.COMMON_NAME, common_name),
        x509.NameAttribute(NameOID.ORGANIZATION_NAME, organization_name),
    ])

    now = datetime.now(timezone.utc)

    # Build self-signed CA certificate
    builder = (
        x509.CertificateBuilder()
        .subject_name(subject)
        .issuer_name(issuer)
        .public_key(ca_key.public_key())
        .serial_number(x509.random_serial_number())
        .not_valid_before(now - timedelta(minutes=1))
        .not_valid_after(now + timedelta(days=days_valid))
        .add_extension(
            x509.BasicConstraints(ca=True, path_length=None),
            critical=True,
        )
        .add_extension(
            x509.KeyUsage(
                digital_signature=True,
                key_encipherment=False,
                content_commitment=False,
                data_encipherment=False,
                key_agreement=False,
                key_cert_sign=True,
                crl_sign=True,
                encipher_only=False,
                decipher_only=False,
            ),
            critical=True,
        )
    )

    # SubjectKeyIdentifier & AuthorityKeyIdentifier
    ski = x509.SubjectKeyIdentifier.from_public_key(ca_key.public_key())
    builder = builder.add_extension(ski, critical=False)
    builder = builder.add_extension(
        x509.AuthorityKeyIdentifier(
            key_identifier=ski.digest,
            authority_cert_issuer=None,
            authority_cert_serial_number=None,
        ),
        critical=False,
    )

    ca_cert = builder.sign(
        private_key=ca_key,
        algorithm=hashes.SHA256(),
    )

    return (
        ca_key,
        ca_cert,
    )


def generate_tls(
    *,
    ca_key: PrivateKey,
    ca_cert: x509.Certificate,
    common_name: str,
    is_client: bool = True,
    is_server: bool = False,
    days_valid: int = DEFAULT_DAYS_VALID,
) -> Tuple[
    PrivateKey,
    x509.Certificate,
]:
    tls_priv = generate_private_key(type="rsa")
    tls_pub = tls_priv.public_key()

    # Subject for this cert
    subject = x509.Name([
        x509.NameAttribute(NameOID.COMMON_NAME, common_name),
    ])

    issuer = ca_cert.subject

    now = datetime.now(timezone.utc)
    builder = (
        x509.CertificateBuilder()
        .subject_name(subject)
        .issuer_name(issuer)
        .public_key(tls_pub)
        .serial_number(x509.random_serial_number())
        .not_valid_before(now - timedelta(minutes=1))
        .not_valid_after(now + timedelta(days=days_valid))
        .add_extension(
            x509.BasicConstraints(ca=False, path_length=None),
            critical=True,
        )
    )

    # Extended key usages for TLS client / server
    eku_usages: List[ObjectIdentifier] = []
    if is_client:
        eku_usages.append(ExtendedKeyUsageOID.CLIENT_AUTH)
    if is_server:
        eku_usages.append(ExtendedKeyUsageOID.SERVER_AUTH)

    if eku_usages:
        builder = builder.add_extension(
            x509.ExtendedKeyUsage(eku_usages),
            critical=False,
        )

    # Sign with private key
    tls_cert = builder.sign(
        private_key=ca_key,
        algorithm=hashes.SHA256(),
    )

    return (
        tls_priv,
        tls_cert,
    )


def get_public_key_as_string(
    certificate: x509.Certificate,
) -> str:
    bytes = certificate.public_key().public_bytes(
        encoding=Encoding.DER,
        format=PublicFormat.SubjectPublicKeyInfo,
    )

    return b64encode(bytes).decode("ascii")



from typing import Literal, Union, overload

from cryptography.hazmat.primitives.asymmetric import ec, rsa

__all__ = [
    "DEFAULT_RSA_KEY_SIZE",
    "PrivateKeyType",
    "PrivateKey",
    "generate",
]

DEFAULT_RSA_KEY_SIZE = 2048

PrivateKeyType = Literal["rsa", "ecdsa"]
PrivateKey = Union[rsa.RSAPrivateKey, ec.EllipticCurvePrivateKey]


@overload
def generate(
    *,
    type: Literal["rsa"],
    key_size: int = 2048,
) -> rsa.RSAPrivateKey:
    ...


@overload
def generate(
    *,
    type: Literal["ecdsa"],
) -> ec.EllipticCurvePrivateKey:
    ...


def generate(
    *,
    type: PrivateKeyType,
    key_size: int = DEFAULT_RSA_KEY_SIZE,
):
    if type == "rsa":
        return rsa.generate_private_key(
            public_exponent=65537,
            key_size=key_size,
        )

    elif type == "ecdsa":
        return ec.generate_private_key(
            curve=ec.SECP256R1(),
        )

    else:
        raise ValueError(f"unknown type: {type}")



#!/usr/bin/env python3

import os
from typing import Any

from setuptools import find_packages, setup

package = "crunch_certificate"


about: dict[str, Any] = {}
here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, package, "__version__.py")) as f:
    exec(f.read(), about)

with open("requirements/default.txt") as fd:
    requirements = fd.read().splitlines()

with open("requirements/test.txt") as fd:
    test_requirements = fd.read().splitlines()

with open("README.md") as fd:
    readme = fd.read()

setup(
    name=about["__title__"],
    description=about["__description__"],
    long_description=readme,
    long_description_content_type="text/markdown",
    version=about["__version__"],
    author=about["__author__"],
    author_email=about["__author_email__"],
    url=about["__url__"],
    packages=find_packages(),
    package_data={
        package: [
            "web_sign.html",
        ]
    },
    python_requires=">=3",
    install_requires=requirements,
    extras_require={
        "test": test_requirements,
    },
    zip_safe=False,
    entry_points={
        "console_scripts": [
            "crunch-certificate=crunch_certificate.cli:cli",
            "crunch-cert=crunch_certificate.cli:cli",
        ],
    },
    classifiers=[
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3.7",
    ],
    keywords="package development template"
)




from typing import Optional, cast, get_args, overload

from cryptography import x509
from cryptography.hazmat.primitives import serialization

from crunch_certificate.private_key import PrivateKey


@overload
def dumps(
    *,
    private_key: PrivateKey,
) -> str:
    ...


@overload
def dumps(
    *,
    certificate: x509.Certificate,
) -> str:
    ...


def dumps(
    *,
    private_key: Optional[PrivateKey] = None,
    certificate: Optional[x509.Certificate] = None,
) -> str:
    if private_key is not None:
        return (
            private_key
            .private_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PrivateFormat.PKCS8,
                encryption_algorithm=serialization.NoEncryption(),
            )
            .decode()
        )

    elif certificate is not None:
        return (
            certificate
            .public_bytes(serialization.Encoding.PEM)
            .decode()
        )

    else:
        raise TypeError("nothing to stringify")


def loads_certificate(
    pem_string: str,
) -> x509.Certificate:
    return x509.load_pem_x509_certificate(
        pem_string.encode(),
    )


def loads_private_key(
    pem_string: str,
) -> PrivateKey:
    private_key = serialization.load_pem_private_key(
        pem_string.encode(),
        password=None,
    )

    if not isinstance(private_key, get_args(PrivateKey)):
        raise ValueError(f"unsupported key: {type(private_key)}")

    return cast(PrivateKey, private_key)
